# encoding: UTF-8
import os.path
from copy import copy
import shelve
import signal

vars = dict()
varnames = "xyztspqfgabcdehijklmnoruvwABCDEFGHIJKLMNOPQRSTUVWXYZ"
DBFILE = os.path.expanduser("~/.config/pylambda.vars")


def lambda_pow(var,expr,const):
  "Returns right part of (\var.expr^{n})"
  if const.num == 0:
    return var
  else:
    f = lambda_pow(var,expr,Const(const.num-1))
    return expr.subst(var.show(),f)

class Expr(object):
  "Generic expression."
  number = 0
  def __init__(self,op,a1,a2):
    "_op_ is an operation (+,-,*,/,...), a1 and a2 are Expr objects."
    Expr.number += 1
    self.number = Expr.number
    self.op = op
    self.a1 = a1
    self.a2 = a2

  def show(self):
    "String representation of an object."
    if self.op == "@":
      return "(%s %s)" % (self.a1.show(),self.a2.show())
    else:
      return "(%s %s %s)" % (self.a1.show(),self.op,self.a2.show())

  def subst(self, old,new):
    """Substitute _old_ for _new_ in this Expr.
    old must be a string; new must be an Expr."""
    return Expr(self.op,self.a1.subst(old,new), self.a2.subst(old,new))

  def lambda_vars(self):
    return self.a1.lambda_vars() + self.a2.lambda_vars()

  def freeof(self,name):
    return self.a1.freeof(name) and self.a2.freeof(name)

  def alpha(self,badlist,vlist=[]):
#     print ">> [%d]Expr.Renaming off %s" % (self.number,badlist)
    a = self.a1.alpha(badlist,vlist)
    b = self.a2.alpha(badlist,vlist)
    return Expr(self.op,a,b)

  def apply(self,expr,excl_list=[]):
    "Try apply self (as a function) to given Expr."
#     print ">> [%d]Expr.Apply: <%s> to <%s>" % (self.number,self.show(),expr.show())
    t = self.eval(excl_list)
    if isinstance(t,Lambda):
      return t.apply(expr.alpha(excl_list),excl_list)
    else:
      return Expr("@",self,expr)

  def eval_vars(self,excl_list=[]):
    "Evaluate expression and substitute variables values where possible."
#     print ">> [%d]Expr.Eval_vars: <%s> excl_list = %s" % (self.number,self.show(),excl_list)
    tmp = self.eval(excl_list)
    if isinstance(tmp,Lambda) or isinstance(tmp,Var) or isinstance(tmp,Const) or isinstance(tmp,Null):
      return tmp
    else:
      self = tmp
    if isinstance(self.a1,Const) and isinstance(self.a2,Const):
      # Perform arithmetics.
      a = self.a1
      b = self.a2
      if self.op == '+':
        return Const(a.num+b.num)
      elif self.op == '-':
        return Const(a.num-b.num)
      elif self.op == '*':
        return Const(a.num*b.num)
      elif self.op == '/':
        try:
          return Const(a.num/b.num)
        except ZeroDivisionError:
          print "Division by zero."
          return Null("failed")
      elif self.op == '^':
        return Const(a.num ** b.num)
      elif self.op == '!':
        print "Cannot function-style exponent for constant."
        return Null()
      else:
        print "Should not reach: unknown operation."
    if self.op == "!":
      t = self.a1.eval_vars(excl_list)
      if isinstance(self.a2,Const) and isinstance(t,Lambda):
        return Lambda(t.var,lambda_pow(t.var,t.expr,self.a2))
    if self.op == "@":
      t = self.a1.eval_vars(excl_list)
      return t.apply(self.a2.eval_vars(excl_list).alpha(excl_list),excl_list)
    return Expr(self.op,self.a1.eval_vars(excl_list),self.a2.eval_vars(excl_list).alpha(excl_list))

  def simplify(self):
    "Try simplifying this Expr. Does not work yet."
    if self.op == '+':
      if self.a1.show() == self.a2.show():
        return Expr("*",Const(2),self.a1)
      if isinstance(self.a2,Expr) and self.a2.op == "*" and isinstance(self.a2.a1,Const) and self.a1.show() == self.a2.a2.show():
        return Expr("*",Const(self.a2.a1.num+1),self.a1)
    elif self.op == '-':
      if self.a1.show() == self.a2.show():
        return Expr("+",Const(0),Const(0))
    return self

  def eval(self,excl_list=[]):
    "Evaluate Expr."
#     print ">> [%d]Expr.Eval: <%s> excl_list = %s" % (self.number,self.show(), excl_list)
    lt = self.a1.eval(excl_list)
    if self.op == "@":
      if isinstance(lt,Lambda):
#       if isinstance(self.a2,Var) and isinstance(lt,Lambda):
        lt = lt.alpha([self.a2.show()])
#       print ">> [%d]Expr.Eval: LT = <%s>" % (self.number,lt.show())
      l = lt.lambda_vars()
      e = lt.apply(self.a2.alpha(excl_list,l),excl_list)
      return  e
    elif self.op == "!":
      if isinstance(lt,Lambda) and isinstance(self.a2,Const):
        e = Lambda(lt.var,lambda_pow(lt.var,lt.expr,self.a2))
      else:
        e = Expr(self.op,lt,self.a2.eval(excl_list))
      return e
    else: # +,-,*,/,^
      if isinstance(lt,Lambda) and isinstance(self.a2,Lambda):
        return Expr(self.op,lt,self.a2)
      if isinstance(lt,Lambda):
        return Lambda(lt.var,Expr(self.op,lt.expr,self.a2))
      elif isinstance(self.a2,Lambda):
        return Lambda(self.a2.var,Expr(self.op,lt,self.a2.expr))
      else:
        e = Expr(self.op,lt,self.a2.eval(excl_list))
        return e

class Assignment(Expr):
  "Assignment, as in x = 5"
  def __init__(self,var,expr):
    self.var = var
    self.expr = expr
    vars[var.show()] = expr

  def show(self):
    return "%s = %s" % (self.var.show(),self.expr.show())

  def alpha(self,badlist,varlist=[]):
    return Assignment(self.var.alpha(badlist,varlist), self.expr.alpha(badlist,varlist))

  def eval_vars(self,excl_list=[]):
#     print ">> Eval_vars: <%s> excl_list = %s" % (self.show(),excl_list)
    return self

  def eval(self,excl_list=[]):
    return self

class Comma(Expr):
  "Variable substitution expression, as in: 'sqr z, z=5'."
  def __init__(self,expr,assign):
    self.expr = expr
    self.assign = assign

  def show(self):
    return "%s, %s" % (self.expr.show(),self.assign.show())

  def subst(self,old,new):
    e = self.expr.subst(old,new)
    return Comma(e,self.assign)

  def eval_vars(self,excl_list=[]):
#     print ">> Eval_vars: <%s> excl_list = %s" % (self.show(),excl_list)
    t = self.expr.eval_vars(excl_list)
    return t.subst(self.assign.var.show(),self.assign.expr).eval_vars(excl_list)

  def eval(self,excl_list=[]):
    return self.expr.subst(self.assign.var.show(),self.assign.expr).eval(excl_list)
    
class Null(Expr):
  def __init__(self,val=""):
    self.val = val
    pass
  def show(self):
    return self.val
  def lambda_vars(self):
    return []
  def eval_vars(self,excl_list=[]):
#     print ">> Eval_vars: <%s> excl_list = %s" % (self.show(),excl_list)
    return self
  def eval(self,excl_list=[]):
    return self

def multy_lambda(arglist,expr):
  "Return Lambda(arglist[0],Lambda(arglist[1],...Lambda(arglist[n],expr))..)."
  if len(arglist) == 1:
    return Lambda(Var(arglist[0]),expr)
  else:
    return Lambda(Var(arglist[0]),multy_lambda(arglist[1:],expr))

def first_var(num,excl_list=[]):
  n = num
  while varnames[n] in excl_list:
    n = n+1
  return varnames[n]

def nextvar(old):
  if old.count("'") < 3:
    n = old + "'"
  else:
    o = old[0]
    i = (varnames.index(o)+1)%len(varnames)
    n = varnames[i]
  return n

class Lambda(Expr):
  "Lambda-expression."

  number = 0

  def __init__(self,var,expr,rename=True,excl_list=[]):
    Lambda.number += 1
    self.number = Lambda.number
    if isinstance(var,Lambda):
      print "WTF?"
    self.var = var
    self.expr = expr

  def show(self):
    return "(λ%s.%s)" % (self.var.show(), self.expr.show())

  def lambda_vars(self):
    return [self.var.show()]+self.expr.lambda_vars()

  def freeof(self,name):
    if self.var.show() == name:
      return True
    return self.expr.freeof(name)

  def alpha(self, badlist, vlist = []):
#     print ">> [%d]Lambda.Renaming <%s> off %s." % (self.number,self.show(),badlist)
    o = self.var.show() 
    s = nextvar(o)
    while s in badlist:
      s = nextvar(s)
    b = copy(vlist)
    if o in badlist:
      v = Var(s)
#       print ">> [%d]Lambda.Rename: renamed <%s> to <%s>." % (self.number,o,s)
      b.append(o)
      e = self.expr.alpha(badlist,b)
      r = Lambda(v,e,False)
    else:
#       print ">> [%d]Lambda.Rename: will not rename var: my var is <%s>" % (self.number, o)
      e = self.expr.alpha(badlist,b)
#       e = self.expr
      r = Lambda(self.var,e,False)
#     print ">> [%d]Lambda.Rename result = <%s>" % (self.number,r.show())
    return r

  def subst(self,old,new):
    if self.var.show() != old:
      return Lambda(self.var,self.expr.subst(old,new))
    else:
      return self

  def apply(self,expr,excl_list=[]):
#     print ">> [%d]Lambda.Apply Excl_list = %s." % (self.number,excl_list)
    l = self.lambda_vars()
#     print ">> [%d]Lambda.Apply Lambda_vars = %s" % (self.number,l)
    q = expr.alpha(l)
    e = self.expr.subst(self.var.show(),q)
#     print ">> [%d]Lambda.Apply <%s> to <%s> result: <%s>" % (self.number,self.show(),q.show(),e.show())
    return e

  def eval_vars(self,excl_list=[]):
#     print ">> Lambda.Eval_vars: <%s> excl_list = %s" % (self.show(),excl_list)
    l = copy(excl_list)
    l.append(self.var.show())
    e = self.expr.eval_vars(l).alpha(l)
    return Lambda(self.var,e)

  def eval(self,excl_list=[]):
#     print ">> [%d]Lambda.Eval: <%s> excl_list = %s" % (self.number,self.show(),excl_list)
    if self.var.show() in excl_list:
      t = self.alpha(excl_list)
    else:
      t = copy(self)
    l = copy(excl_list)
    l.append(t.var.show())
#     print " > > [ % d]Lambda.Eval: L = % s" % (self.number, l)
    # Perform η-reduction
    try:
      if t.expr.op == "@" and isinstance(t.expr.a2,Var) and not isinstance(t.expr.a2,Const):
        if t.var.show() == t.expr.a2.show() and t.expr.a1.freeof(t.var.show()):
          return t.expr.a1
    except:
      pass
    return Lambda(t.var,t.expr.eval(l))

def is_lambda(something):
  if isinstance(something,Lambda):
    return True
  elif isinstance(something,Var):
    return typeof(something)=='Lambda'
  else:
    return False

def typeof(var):
  v = var
  while isinstance(v,Var) and not isinstance(v,Const):
    try:
      v = vars[v.name]
    except KeyError:
      return 'UnboundVar'
  if isinstance(v,Const):
    return 'Const'
  elif isinstance(v,Lambda):
    return 'Lambda'
  elif isinstance(v,Assignment):
    return 'Assignment'
  elif isinstance(v,Comma):
    return 'Comma-Expr'
  elif isinstance(v,Expr):
    return 'Expr'

class Var(Expr):
  "A variable."
  def __init__(self,name):
    self.name = name

  def show(self):
    return self.name

  def subst(self,old,new):
    if self.name == old:
      return copy(new)
    else:
      return copy(self)

  def freeof(self,name):
    return (self.name != name)

  def lambda_vars(self):
    return []

  def alpha(self,badlist,vlist=[]):
#     return self
#     print ">> [%s]Var.Renaming off %s, %s" % (self.name,badlist,vlist)
    if self.name in vlist and not self.name in vars:
      o = nextvar(self.name)
      while o in vlist+badlist:
        o = nextvar(o)
      v = Var(o)
#       print ">> [%s]Var.Rename: Lists: %s, %s. Renaming to <%s>" % (self.name,badlist,vlist,o)
      return v
    else:
      return self

  def apply(self,expr,excl_list=[]):
#     print ">> Var.Apply called."
    return Expr('@',copy(self),expr)

  def eval_vars(self,excl_list=[]):
#     print ">> [%s]Var.Eval_vars: excl_list = %s" % (self.show(),excl_list)
    if self.name in excl_list:
      return Var(self.name)
    else:
      e = self.eval(excl_list).alpha(excl_list)
#       print ">> Eval_vars: Var <%s> evaluated to <%s>." % (self.name,e.show())
      return e

  def eval(self,excl_list=[]):
#     print ">> [%s]Var.Eval: excl_list = %s" % (self.show(),excl_list)
    if (self.name in vars) and not (self.name in excl_list):
      e = vars[self.name].alpha(excl_list)
#       print ">> [%s]Var.Eval: read <%s> from hash" % (self.show(),e.show())
      if isinstance(e,Lambda):
        if e.var.show() in excl_list:
#           print ">> [%s]Var.Eval: this is Lambda, creating new Lambda with new var." % self.show()
          e = Lambda(e.var,e.expr.alpha(excl_list),True,excl_list)
#           print ">> -------"
    else:
      e = copy(self)
#     print ">> Var.Eval: Var <%s> evaluated to <%s>." % (self.name,e.show())
    return e

class Const(Var):
  "Integer constant."
  def __init__(self,num):
    self.num = num

  def show(self):
    return str(self.num)

  def freeof(self,name):
    return True

  def alpha(self,badlist,vlist=[]):
    return self

  def subst(self,old,new):
    return copy(self)

  def eval_vars(self,excl_list=[]):
#     print ">> Eval_vars: <%s> excl_list = %s" % (self.show(),excl_list)
    return copy(self)

  def apply(self,expr,excl_list=[]):
    print "Cannot apply number as a function!"
    return Null("failed")

  def eval(self,excl_list=[]):
    return copy(self)

def merge(d1,d2):
  t = copy(d1)
  for k in d2:
    t[k] = d2[k]
  return t

def set_old_value(data):
  global vars
  vars['%'] = data

def show_var(name):
  global vars
  return "%s = %s" % (name,vars[name].show())

def show_vars():
  global vars
  print ", ".join(map(show_var,vars.keys()))

def save():
  global vars
  d = shelve.open(DBFILE)
  tmp = copy(vars)
  del vars['%']
  d["vars"] = tmp
  d.close()

def load():
  global vars
  try:
    old = copy(vars['%'])
  except KeyError:
    old = Null("empty")
  d = shelve.open(DBFILE)
  t = d["vars"]
  vars = merge(vars,t)
  d.close()
  vars['%'] = old

def catch_signal(signum,frame):
  raise RuntimeError, "User interrupt."
#   Expr.stop = 1

