#!/usr/bin/python

import sys
import readline
import signal
from ply import yacc

from lambdalex import tokens
# from data import Assignment,Comma,Expr,Lambda,Var,Const,multy_lambda,typeof,load,save,show_vars
from data import *

precedence = (
  ('left','EQ'),
  ('left','COMMA'),
  ('left','SPACE'),
  ('left','APPLY'),
  ('left','ADDOP','SUBOP'),
  ('left','MULOP','DIVOP'),
  ('left','FUNCPOW'),
  ('left','POWOP'))

def p_error(t):
  raise TypeError("Syntax error at <%s>!" % t.value)

def p_generic_expr(t):
  'generic : expr'
  t[0] = t[1]

def p_generic_assign(t):
  'generic : assignment'
  t[0] = t[1]

def p_assignment(t):
  'assignment : VAR EQ expr'
  if not t[3].freeof(t[1]):
    print "Warning: recursive assignment."
  t[0]= Assignment(Var(t[1]),t[3])

def p_expr_plus(t):
  'expr : expr ADDOP expr'
  t[0] = Expr('+',t[1],t[3])

def p_expr_minus(t):
  'expr : expr SUBOP expr'
  t[0] = Expr('-',t[1],t[3])

def p_expr_mul(t):
  'expr : expr MULOP expr'
  t[0] = Expr('*',t[1],t[3])

def p_expr_div(t):
  'expr : expr DIVOP expr'
  t[0] = Expr('/',t[1],t[3])

def p_expr_pow(t):
  'expr : expr POWOP expr'
  t[0] = Expr('^',t[1],t[3])

def p_expr_funcpow(t):
  'expr : expr FUNCPOW expr'
  t[0] = Expr('!',t[1],t[3])

def p_expr_var(t):
  'expr : VAR'
  t[0] = Var(t[1])

def p_expr_number(t):
  'expr : NUMBER'
  t[0] = Const(t[1])

def p_expr_paren(t):
  'expr : BRACKETL expr BRACKETR'
  t[0] = t[2]

def p_expr_comma(t):
  'expr : expr COMMA assignment'
  t[0] = Comma(t[1],t[3])

def p_expr_apply(t):
  'expr : expr APPLY expr'
  t[0] = Expr('@',t[1],t[3])

def p_expr_apply_space(t):
  'expr : expr SPACE expr'
  t[0] = Expr('@',t[1],t[3])

# def p_arglist_dot(t):
#   'arglist : argdotlist'
#   t[0] = t[1]

# def p_arglist_space(t):
#   'arglist : argspacelist'
#   t[0] = t[1]
# 
# def p_argdotlist(t):
#   'argdotlist : argdotlist DOT VAR'
#   q = t[1]
#   q.append(t[3])
#   t[0] = q

def p_argspacelist(t):
  'argspacelist : argspacelist SPACE VAR'
  q = t[1]
  q.append(t[3])
  t[0] = q

# def p_argdotlist_var(t):
#   'argdotlist : VAR'
#   t[0] = [t[1]]

def p_argspacelist_var(t):
  'argspacelist : VAR'
  t[0] = [t[1]]

def p_expr_lambda(t):
  'expr : lambda'
  t[0] = t[1]

# def p_lambda(t):
#   'lambda : BRACKETL LAMBDA VAR DOT expr BRACKETR'
#   t[0] = Lambda(Var(t[3]),t[5])

# def p_lambda(t):
#   'lambda : LAMBDA VAR DOT expr'
#   t[0] = Lambda(Var(t[2]),t[4],True)

def p_lambda(t):
  'lambda : LAMBDA argspacelist DOT expr'
  t[0] = multy_lambda(t[2],t[4])

def read():
  s = raw_input('? ')
  s = s.strip()
  if s == "exit":
    sys.exit()
  elif s == "save":
    save()
    print "Variables saved."
    return ""
  elif s == "load":
    load()
    print "Variables loaded."
    return ""
  elif s == "vars":
    show_vars()
    return ""
  return s

if __name__ == "__main__":
  signal.signal(signal.SIGINT,catch_signal)
  yacc.yacc()
  while True:
    try:
      s = read()
    except RuntimeError:
      print ""
      continue
    except EOFError:
      print "exiting."
      break
    if s == "" : continue
    try:
      value = yacc.parse(s).eval().eval_vars()
      print ": %s  :: %s" % (value.show(), typeof(value))
      set_old_value(value)
    except TypeError,e:
      print e
    except RuntimeError, e:
      print e
    Lambda.varnum=0
    counter = 0


